#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <time.h>
#include <sys/asoundlib.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include "../wmgeneral/wmgeneral.h"
#include "../wmgeneral/misc.h"
#include "master.xpm"

char wmix_mask_bits[64*64];
int  wmix_mask_width = 64;
int  wmix_mask_height = 64;

// X11 stuff
XWindowAttributes attr;
Window Root;
GC geecee;
XGCValues gcv;
unsigned long gcm;

struct {
 int screen;
 int width;
 int height;
 int depth;
 int font;
 int num_channels;
 int showbar;
 int channel;
 int sel;
 int osd;
 int dragging;
 char *colorant;
} wmix;

int curpos[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int preset[]={0,0,0,0,0,0,0};
static int card_id = 0;
static int mixer_id = 0;    
static void *mixer_handle;
static char *mixer_card_name = NULL;
static char *mixer_device_name = NULL;
snd_mixer_channel_info_t cinfo;
snd_mixer_channel_t cdata = { 0 };

#define WMIX_VERSION "1.5"

int update_rate=200000;

void usage(void);
void printversion(void);
void wmix_routine(int, char **);
void drawinfo(int);
char *mid(char *, int, int);
void pressEvent(XButtonEvent *hmm);
void releaseEvent(XButtonEvent *hmm);
void motionEvent(XMotionEvent *hmm);
void draw_slider(int);
void draw_numerics(int);
void xpmprint(char *, int, int);
void scroll(char *,int delay);
void draw_text(char *, int, int);
unsigned long getColor(char *);
char *channel_name(int);

unsigned long getColor(char *colorname){
 XColor color;
 XWindowAttributes winattr;
 XGetWindowAttributes(display, Root, &winattr);
 color.pixel=0;
 XParseColor(display, winattr.colormap, colorname, &color);
 color.flags=DoRed | DoGreen | DoBlue;
 XAllocColor(display, winattr.colormap, &color);
 return color.pixel;
}

void draw_text(char *text, int x, int y)
{
 XDrawString(display,Root,geecee,x,y,text,strlen(text));
}
                        
int mixer_init (void)
{
 static snd_mixer_info_t mixer_info = { 0 };
 static struct snd_ctl_hw_info hw_info;
 void *ctl_handle;

 if (snd_ctl_open( &ctl_handle, card_id ) < 0 ) fprintf(stderr,"snd_ctl_open sucked.\n");
 if (snd_ctl_hw_info( ctl_handle, &hw_info ) < 0 ) fprintf(stderr,"snd_ctl_hw_info sucked.\n" );
 snd_ctl_close( ctl_handle );

 if (snd_mixer_open( &mixer_handle, card_id, mixer_id ) < 0) fprintf(stderr,"snd_mixer_open sucked.\n");

 if (snd_mixer_info( mixer_handle, &mixer_info) < 0)
 {
  fprintf(stderr,"snd_mixer_info sucked.\n");
  fprintf(stderr,"With all these errors, you probably dont have ALSA installed, dipshit.\n");
  return 0;
 }
 wmix.num_channels = mixer_info.channels;
 mixer_card_name = hw_info.name;
 mixer_device_name = mixer_info.name;
 return 1;
}

char *channel_name(int n)
{
 snd_mixer_channel_info(mixer_handle,n,&cinfo);
 return cinfo.name;
}
 
int main(int argc, char *argv[])
{
 int		i;
 FILE *fe;
 char mixfile[256];

 for (i=1; i<argc; i++) {
  char *arg = argv[i];
  if (*arg=='-')
  {
   switch (arg[1]) {
    case 'c' :
     if (argv[i+1]) card_id=atoi(argv[i+1]);
    break;
    case 'r' :
     wmix.colorant=argv[i+1];
     break;
    case 'v' :
     printversion();
     exit(0);
    break;
    default:
     usage();
     exit(0);
    break;
   }
  }
 }

 if (!mixer_init()) 
  { fprintf(stderr,"Segmentation fault, core dumped.\n"); exit(1); }

 for (i=0; i<=wmix.num_channels; i++) {
  snd_mixer_channel_read(mixer_handle, i, &cdata);
  curpos[i]=(cdata.left+cdata.right)/2;
 }

 sprintf(mixfile,"%s/.wmixrc",getenv("HOME"));
 fe=fopen(mixfile,"r");
 for(i=0;i<8;i++) {
  fscanf(fe,"%d\n",&preset[i]);
 }
 fclose(fe);

 wmix.channel=0;
 wmix.sel=0;
 wmix.osd=0;
 wmix.dragging=0;
 if (wmix.colorant==NULL) wmix.colorant="green";
 wmix_routine(argc, argv);
 return 0;
}

void xpmprint(char *text, int x, int y)
{
 int i;
 int c;
 int k;

 k = x;
 copyXPMArea(73,67,57,9,x,y);
 if (strlen(text)>8) text[9]='\0';

 for (i=0; text[i]; i++)
 {
  c = toupper(text[i]); 
  if (c=='-') { copyXPMArea(60,67,6,8,k,y); k += 6; }
  if (c==' ') { copyXPMArea(66,67,6,8,k,y); k += 6; }
  if (c >= 'A' && c <= 'Z') // letter
  {
   c = c - 'A';
   copyXPMArea(c * 6, 77, 6, 8, k, y);
   k += 6;
  } else if (c >= '0' && c <= '9') // number
   {
    c = c - '0';
    copyXPMArea(c * 6, 67, 6, 8, k, y);
    k += 6;
   }
 }
}

char *mid(char *text, int offset, int size)
{
 char *tmp;
 tmp=malloc(sizeof(char)*30);
 text+=offset;
 return strncpy(tmp,text,size);
}

void scroll(char *scrolltext,int delay)
{
 int i;
 for(i=0;i<strlen(scrolltext);i++) {
  xpmprint(mid(scrolltext,i,9),4,15);RedrawWindow();usleep(delay*100000);
 }
}

void drawinfo(int on)
{
  int i;
  char *tmp;

  for (i=0; i<=6; i++) {
     if (on==i) {
      copyXPMArea(67,55,8,9,(i*8)+2,3);
      if (wmix.sel==1) {
       FILE *fe;
       char mixfile[256];
       sprintf(mixfile,"%s/.wmixrc",getenv("HOME"));
       preset[i]=wmix.channel;
       scroll("        SAVED",1);
       wmix.showbar=0;
       fe=fopen(mixfile,"w");
       for(i=0;i<8;i++) {
        fprintf(fe,"%d\n",preset[i]);
       }
       fclose(fe);
       wmix.sel=0;
      } else {
       wmix.channel=preset[i];
      }
     } else copyXPMArea(76,55,8,9,(i*8)+2,3);
  }

  switch (on) {
   case 7: scroll("        PHEAR THE MIGHTY TROUTZ  TIMECOP",2); wmix.showbar=0; break;

   case 8:
    wmix.sel=1; wmix.showbar=0;
    scroll("        CHOOSE A PRESET",1);
   break;

   case 9: 
    wmix.channel--;
    if (wmix.channel<0) wmix.channel=wmix.num_channels-1;
   break;

   case 10:
    wmix.channel++;
    if (wmix.channel>wmix.num_channels-1) wmix.channel=0;
   break;

   case 11:
    snd_mixer_channel_info(mixer_handle, wmix.channel, &cinfo);
    if (cinfo.caps & SND_MIXER_CINFO_CAP_RECORD) {
     snd_mixer_channel_read(mixer_handle, wmix.channel, &cdata);
     if (cdata.flags & SND_MIXER_FLG_RECORD) {
      copyXPMArea(81,30,7,10,4,29);
      cdata.flags &= ~SND_MIXER_FLG_RECORD;
     } else {
      copyXPMArea(74,30,7,10,4,29);
      cdata.flags |= SND_MIXER_FLG_RECORD;
     }
     snd_mixer_channel_write(mixer_handle, wmix.channel, &cdata);
    }
    wmix.showbar=0;
   break;

   case 12:
    snd_mixer_channel_info(mixer_handle, wmix.channel, &cinfo);
    if (cinfo.caps & SND_MIXER_CINFO_CAP_MUTE) {
     if (cdata.flags & SND_MIXER_FLG_MUTE)
      cdata.flags &= ~SND_MIXER_FLG_MUTE;
     else
      cdata.flags |= SND_MIXER_FLG_MUTE;
     snd_mixer_channel_write(mixer_handle, wmix.channel, &cdata);
    }
    wmix.showbar=0;
   break;

   case 13:
    tmp=malloc(sizeof(char)*(strlen(mixer_card_name)+10));
    strcat(tmp,"        ");
    strcat(tmp,mixer_card_name);
    scroll(tmp,2);
    wmix.showbar=0;
    free(tmp);
   break;

  }  // case end
     
  snd_mixer_channel_info(mixer_handle, wmix.channel, &cinfo);
  if (cinfo.caps & SND_MIXER_CINFO_CAP_RECORD) {
   snd_mixer_channel_read(mixer_handle, wmix.channel, &cdata);
   if (cdata.flags & SND_MIXER_FLG_RECORD)
    copyXPMArea(74,30,7,10,4,29);
   else
    copyXPMArea(81,30,7,10,4,29);
  } else
   copyXPMArea(67,30,7,10,4,29);  

  snd_mixer_channel_info(mixer_handle, wmix.channel, &cinfo);
   if (cinfo.caps & SND_MIXER_CINFO_CAP_MUTE) {
    snd_mixer_channel_read(mixer_handle, wmix.channel, &cdata);
    if (cdata.flags & SND_MIXER_FLG_MUTE)
     copyXPMArea(88,30,7,10,4,42);
    else
     copyXPMArea(67,30,7,10,4,42);
   }

  if (on==-1) on=0;
  draw_slider(54-(curpos[wmix.channel]/5));
  xpmprint(channel_name(wmix.channel),4,15);
  RedrawWindow();
}

void pressEvent(XButtonEvent *hmm)
{
 int x=hmm->x;
 int y=hmm->y;
 
 if(x>=19 && x<=40 && y>=34 && y<=54)
 {
  wmix.dragging=1;
  draw_slider(y);
 }
}
      
void releaseEvent(XButtonEvent *hmm)
{
 wmix.dragging=0;
 RedrawWindow();
}

void motionEvent(XMotionEvent *hmm)
{
 int x=hmm->x;
 int y=hmm->y;
 if (x>=2 && y>=27 && x<=42 && y<=61 && wmix.dragging)
 { 
  if(x>=12 && x<=33 && y>=34 && y<=54) {
   // 28 to 49, 21 total travel time.
   draw_slider(y);
  }
 }
}

void draw_slider(int y)
{
 int i;
 curpos[wmix.channel]=(21-(y-33))*5;
 draw_numerics(curpos[wmix.channel]);
 copyXPMArea(100,6,21,33,19,28);
 copyXPMArea(69,18,21,11,19,y-5);
 snd_mixer_channel_read(mixer_handle, wmix.channel, &cdata);
 cdata.left=curpos[wmix.channel];
 cdata.right=curpos[wmix.channel];
 snd_mixer_channel_write(mixer_handle, wmix.channel, &cdata);
 RedrawWindow();

 XClearArea(display,Root,128,wmix.height-200,500,140,0);
 if (wmix.showbar==1) {
  wmix.osd=10;
  wmix.showbar=1;
  for (i=0;i<(curpos[wmix.channel]/4);i++) {
   XFillRectangle(display,Root,geecee,(i*20)+128,wmix.height-100,5,40);
  }
  draw_text(channel_name(wmix.channel),128,wmix.height-130);
  XFlush(display);
 }
 wmix.showbar=1;
}

void draw_numerics(int num) // double digit, 0 to 99
{
 if (num<100) {
  copyXPMArea((num/10)*7+68,42,7,11,46,50);
  copyXPMArea((num%10)*7+68,42,7,11,53,50);
 } else {
  copyXPMArea(131,42,7,11,46,50);
  copyXPMArea(131,42,7,11,53,50);
 }
}

void wmix_routine(int argc, char **argv)
{
 int i;
 XEvent Event;
 int but_stat=-1;

 createXBMfromXPM(wmix_mask_bits, master_xpm, wmix_mask_width, wmix_mask_height);
 openXwindow(argc, argv, master_xpm, wmix_mask_bits, wmix_mask_width, wmix_mask_height);

 // 'preset' buttons
 AddMouseRegion(0, 3, 4, 10, 11);
 AddMouseRegion(1, 11, 4, 18, 10);
 AddMouseRegion(2, 19, 4, 26, 10);
 AddMouseRegion(3, 27, 4, 34, 10);
 AddMouseRegion(4, 35, 4, 42, 10);
 AddMouseRegion(5, 43, 4, 50, 10);
 AddMouseRegion(6, 51, 4, 58, 10);
 // red/green bar no use yet
 AddMouseRegion(7, 59, 4, 61, 11);
 // channel swap / save preset buttons
 AddMouseRegion(8, 46, 28, 61, 35);
 AddMouseRegion(9, 46, 36, 53, 46);
 AddMouseRegion(10, 54, 36, 61, 46);
 // Record ...
 AddMouseRegion(11, 4, 29, 11, 39);
 // ... and mute toggle
 AddMouseRegion(12, 4, 42, 11, 52);
 // testing for X drawing
 AddMouseRegion(13, 46, 50, 61, 61);

 wmix.screen=DefaultScreen(display);
 Root=RootWindow(display, wmix.screen);
 XGetWindowAttributes(display, Root, &attr);
 wmix.width=attr.width;
 wmix.height=attr.height;
 wmix.depth=attr.depth;
 wmix.showbar=1;
 gcm = GCForeground | GCBackground | GCGraphicsExposures;
 gcv.foreground = getColor(wmix.colorant);
 gcv.background = WhitePixel(display,wmix.screen);
 gcv.graphics_exposures = 0;
 geecee=XCreateGC(display, Root, gcm, &gcv);
 wmix.font=XLoadFont(display,"-misc-fixed-medium-r-normal-*-50-0-75-75-c-*-iso8859-8");
 XSetFont(display,geecee,wmix.font);
 drawinfo(-1);

 while (1)
 {
  wmix.osd--;
  if (wmix.osd==0)  XClearArea(display,Root,128,wmix.height-200,500,140,0);
  
  while (XPending(display))
  {
   XNextEvent(display, &Event);
   switch (Event.type)
    {
     case Expose:
      RedrawWindow();
     break;
     case DestroyNotify:
      XCloseDisplay(display);
      exit(0);
     break;
     case ButtonPress:
      i = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);
      but_stat = i;
      pressEvent(&Event.xbutton);
     break;
     case ButtonRelease:
      i = CheckMouseRegion(Event.xbutton.x, Event.xbutton.y);
      releaseEvent(&Event.xbutton);
      if (but_stat == i && but_stat >= 0) drawinfo(but_stat);
      but_stat = -1;
     break;
     case MotionNotify:
      motionEvent(&Event.xmotion);
     break;
    }
   }
  usleep(update_rate);
 }
}

void usage(void)
{
 fprintf(stderr, "\nWMix - Yet another Window Maker Mixer Applet\nby TimeCop [timecop@linuxwarez.com]\n\n");
 fprintf(stderr, "usage:\n");
 fprintf(stderr, "    -c <card>                 use the mixer on a specific card\n");
 fprintf(stderr, "    -r <color>                draw OSD using this color.\n");
 fprintf(stderr, "    -h                        this help screen\n");
 fprintf(stderr, "    -v                        print the version number\n");
 fprintf(stderr, "\n");
}

void printversion(void)
{
 fprintf(stderr, "wmix v%s\n", WMIX_VERSION);
}
